@@ -249,6 +249,9 @@ usage is void (*_gtk_reserved1)(void);"""
         elif has_element_type:
             self._apply_annotations_element_type(parent, node, options)
 
+        if isinstance(node.type, ast.Array):
+            self._check_array_element_type(node.type, options)
+
     def _resolve(self, type_str, type_node=None, node=None, parent=None):
         def grab_one(type_str, resolver, top_combiner, combiner):
             """Return a complete type, and the trailing string part after it.
@@ -299,6 +302,15 @@ usage is void (*_gtk_reserved1)(void);"""
                               (text, result.ctype), positions=position)
         return result
 
+    def _resolve_toplevel(self, type_str, type_node=None, node=None, parent=None):
+        """Like _resolve(), but attempt to preserve more attributes of original type."""
+        result = self._resolve(type_str, type_node=type_node, node=node, parent=parent)
+        # If we replace a node with a new type (such as an annotated) we
+        # might lose the ctype from the original node.
+        if type_node is not None:
+            result.ctype = type_node.ctype
+        return result
+
     def _get_position(self, func, param):
         block = self._blocks.get(func.symbol)
         if block:
@@ -314,6 +326,29 @@ usage is void (*_gtk_reserved1)(void);"""
 
         return block.position
 
+    def _check_array_element_type(self, array, options):
+        # GPtrArrays are allowed to contain non basic types
+        # (except enums and flags) or basic types that are
+        # as big as a gpointer
+        if array.array_type == ast.Array.GLIB_PTRARRAY and \
+           ((array.element_type in ast.BASIC_GIR_TYPES
+             and not array.element_type in ast.POINTER_TYPES) or
+            isinstance(array.element_type, ast.Enum) or
+            isinstance(array.element_type, ast.Bitfield)):
+            message.warn("invalid (element-type) for a GPtrArray, "
+                        "must be a pointer", options.position)
+
+        # GByteArrays have (element-type) guint8 by default
+        if array.array_type == ast.Array.GLIB_BYTEARRAY:
+            if array.element_type == ast.TYPE_ANY:
+                array.element_type = ast.TYPE_UINT8
+            elif not array.element_type in [ast.TYPE_UINT8,
+                                            ast.TYPE_INT8,
+                                            ast.TYPE_CHAR]:
+                message.warn("invalid (element-type) for a GByteArray, "
+                             "must be one of guint8, gint8 or gchar",
+                             options.position)
+
     def _apply_annotations_array(self, parent, node, options):
         array_opt = options.get(OPT_ARRAY)
         if array_opt:
@@ -495,8 +530,8 @@ usage is void (*_gtk_reserved1)(void);"""
 
         param_type = options.get(OPT_TYPE)
         if param_type:
-            node.type = self._resolve(param_type.one(),
-                                      node.type, node, parent)
+            node.type = self._resolve_toplevel(param_type.one(),
+                                               node.type, node, parent)
 
         caller_allocates = False
         annotated_direction = None
@@ -537,6 +572,7 @@ usage is void (*_gtk_reserved1)(void);"""
         self._adjust_container_type(parent, node, options)
 
         if (OPT_ALLOW_NONE in options or
+            node.type.target_giname == 'Gio.AsyncReadyCallback' or
             node.type.target_giname == 'Gio.Cancellable'):
             node.allow_none = True
 
@@ -722,7 +758,7 @@ usage is void (*_gtk_reserved1)(void);"""
             prop.transfer = self._get_transfer_default(parent, prop)
         type_tag = block.get(TAG_TYPE)
         if type_tag:
-            prop.type = self._resolve(type_tag.value, prop.type, prop, parent)
+            prop.type = self._resolve_toplevel(type_tag.value, prop.type, prop, parent)
 
     def _apply_annotations_signal(self, parent, signal):
         prefix = self._get_annotation_name(parent)
@@ -745,8 +781,8 @@ usage is void (*_gtk_reserved1)(void);"""
                 options = getattr(tag, 'options', {})
                 param_type = options.get(OPT_TYPE)
                 if param_type:
-                    param.type = self._resolve(param_type.one(), param.type,
-                                               param, parent)
+                    param.type = self._resolve_toplevel(param_type.one(), param.type,
+                                                        param, parent)
             else:
                 tag = None
             self._apply_annotations_param(signal, param, tag)
@@ -857,9 +893,7 @@ the ones that failed to resolve removed."""
                 uscore_enums[no_uscore_prefixed] = enum
 
         for node in self._namespace.itervalues():
-            if not isinstance(node, ast.Function):
-                continue
-            if node.retval.type.target_giname != 'GLib.Quark':
+            if not isinstance(node, ast.ErrorQuarkFunction):
                 continue
             short = node.symbol[:-len('_quark')]
             if short == "g_io_error":
@@ -871,7 +905,7 @@ the ones that failed to resolve removed."""
                 if enum is None:
                     enum = uscore_enums.get(short)
             if enum is not None:
-                enum.error_quark = node.symbol
+                enum.error_domain = node.error_domain
             else:
                 message.warn_node(node,
                     """%s: Couldn't find corresponding enumeration""" % (node.symbol, ))
@@ -945,7 +979,7 @@ method or constructor of some type."""
 
         # A quick hack here...in the future we should catch C signature/GI signature
         # mismatches in a general way in finaltransformer
-        if first.type.ctype is not None and first.type.ctype.count('*') != 1:
+        if first.type.ctype is not None and first.type.ctype.count('*') > 1:
             return False
 
         if not func.is_method:
@@ -994,19 +1028,35 @@ method or constructor of some type."""
         if split is None:
             return False
         (node, funcname) = split
-
-        # We actually should treat static methods on a wider class of objects:
-        #  ast.Class, ast.Interface, ast.Record, ast.Union, ast.Boxed
-        # But we stick to ast.Class for now for compatibility with existing code.
-        #
-        # See https://bugzilla.gnome.org/show_bug.cgi?id=572408
-        #
-        if not isinstance(node, ast.Class):
+        if funcname == '':
             return False
 
-        self._namespace.float(func)
-        func.name = funcname
-        node.static_methods.append(func)
+        if isinstance(node, ast.Class):
+            self._namespace.float(func)
+            func.name = funcname
+            node.static_methods.append(func)
+            return True
+        elif isinstance(node, (ast.Interface, ast.Record, ast.Union,
+                               ast.Boxed, ast.Enum, ast.Bitfield)):
+            # prior to the introduction of this part of the code, only
+            # ast.Class could have static methods.  so for backwards
+            # compatibility, instead of removing the func from the namespace,
+            # leave it there and get a copy instead.  modify the copy and push
+            # it onto static_methods.  we need to copy the parameters list
+            # separately, because in the third pass functions are flagged as
+            # 'throws' depending on the presence of a GError parameter which is
+            # then removed from the parameters list.  without the explicit
+            # copy, only one of the two functions would thus get flagged as
+            # 'throws'.  clone() does this for us.
+            new_func = func.clone()
+            new_func.name = funcname
+            node.static_methods.append(new_func)
+            # flag the func as a backwards-comptability kludge (thus it will
+            # get pruned in the introspectable pass if introspectable=0).
+            func.moved_to = node.name + '.' + new_func.name
+            return True
+
+        return False
 
     def _set_up_constructor(self, func, subsymbol):
         self._namespace.float(func)
@@ -1045,15 +1095,23 @@ method or constructor of some type."""
 
         return name
 
+    def _guess_constructor_by_name(self, symbol):
+        # Normal constructors, gtk_button_new etc
+        if symbol.endswith('_new'):
+            return True
+        # Alternative constructor, gtk_button_new_with_label
+        if '_new_' in symbol:
+            return True
+        # gtk_list_store_newv,gtk_tree_store_newv etc
+        if symbol.endswith('_newv'):
+            return True
+        return False
+
     def _is_constructor(self, func, subsymbol):
-        if False and func.symbol == 'regress_constructor':
-            import pdb
-            pdb.set_trace()
         # func.is_constructor will be True if we have a (constructor) annotation
-        if not func.is_constructor and \
-                not (func.symbol.find('_new_') >= 0 or \
-                        func.symbol.endswith('_new')):
-            return False
+        if not func.is_constructor:
+            if not self._guess_constructor_by_name(func.symbol):
+                return False
         target = self._transformer.lookup_typenode(func.retval.type)
         if not (isinstance(target, ast.Class)
                 or (isinstance(target, (ast.Record, ast.Union, ast.Boxed))
